// UBC CPSC 314 (2015W2) -- P1
// HAVE FUN!!! :)

// ASSIGNMENT-SPECIFIC API EXTENSION
THREE.Object3D.prototype.setMatrix = function (a) {
    this.matrix = a;
    this.matrix.decompose(this.position, this.quaternion, this.scale);
}

// SETUP RENDERER & SCENE
var canvas = document.getElementById('canvas');
var scene = new THREE.Scene();
var renderer = new THREE.WebGLRenderer();
renderer.setClearColor(0xFFFFFF); // white background colour
canvas.appendChild(renderer.domElement);

// SETUP CAMERA
var camera = new THREE.PerspectiveCamera(30, 1, 0.1, 1000); // view angle, aspect ratio, near, far
camera.position.set(45, 20, 40);
//camera.position.set(45,0,0);
camera.lookAt(scene.position);
scene.add(camera);

// SETUP ORBIT CONTROLS OF THE CAMERA
var controls = new THREE.OrbitControls(camera);

// ADAPT TO WINDOW RESIZE
function resize() {
    renderer.setSize(window.innerWidth, window.innerHeight);
    camera.aspect = window.innerWidth / window.innerHeight;
    camera.updateProjectionMatrix();
}

// EVENT LISTENER RESIZE
window.addEventListener('resize', resize);
resize();

//SCROLLBAR FUNCTION DISABLE
window.onscroll = function () {
    window.scrollTo(0, 0);
}

// SETUP HELPER GRID
// Note: Press Z to show/hide
var gridGeometry = new THREE.Geometry();
var i;
for (i = -50; i < 51; i += 2) {
    gridGeometry.vertices.push(new THREE.Vector3(i, 0, -50));
    gridGeometry.vertices.push(new THREE.Vector3(i, 0, 50));
    gridGeometry.vertices.push(new THREE.Vector3(-50, 0, i));
    gridGeometry.vertices.push(new THREE.Vector3(50, 0, i));
}

var gridMaterial = new THREE.LineBasicMaterial({color: 0xBBBBBB});
var grid = new THREE.Line(gridGeometry, gridMaterial, THREE.LinePieces);

/////////////////////////////////
//   YOUR WORK STARTS BELOW    //
/////////////////////////////////

var urls = [
    'posx.jpg',
    'negx.jpg',
    'posy.jpg',
    'negy.jpg',
    'posz.jpg',
    'negz.jpg'
];

var cubemap = THREE.ImageUtils.loadTextureCube(urls); // load textures
cubemap.format = THREE.RGBFormat;

var shader = THREE.ShaderLib['cube']; // init cube shader from built-in lib
shader.uniforms['tCube'].value = cubemap; // apply textures to shader

// create shader material
var skyBoxMaterial = new THREE.ShaderMaterial({
    fragmentShader: shader.fragmentShader,
    vertexShader: shader.vertexShader,
    uniforms: shader.uniforms,
    depthWrite: false,
    side: THREE.BackSide
});

// create skybox mesh
var skybox = new THREE.Mesh(
    new THREE.CubeGeometry(1000, 1000, 1000),
    skyBoxMaterial
);

scene.add(skybox);

// MATERIALS
// Note: Feel free to be creative with this!
var bodyMaterial = new THREE.MeshBasicMaterial({color: 0XFFF00A});
var headMaterial = new THREE.MeshBasicMaterial({color: 0X0FFEEA});
var tailMaterial = new THREE.MeshBasicMaterial({color: 0XFF000F});
var noseMaterial = new THREE.MeshBasicMaterial({color: 0XDFE149});
var tentacleMaterial = new THREE.MeshBasicMaterial({color: 0X423DEA});
var pawMaterial = new THREE.MeshBasicMaterial({color: 0XDD3D8F});
var clawMaterial = new THREE.MeshBasicMaterial({color: 0X4F3D82});
var lolMaterial = new THREE.MeshNormalMaterial();

// function drawCube()
// Draws a unit cube centered about the origin.
// Note: You will be using this for all of your geometry
function makeCube() {
    var unitCube = new THREE.BoxGeometry(1, 1, 1);
    return unitCube;
}

function makeCylinder() {
    var unitCyl = new THREE.CylinderGeometry(1, 1, 1, 100);
    return unitCyl;
}

function makeSphere() {
    var unitS = new THREE.SphereGeometry(1);
    return unitS;
}

// GEOMETRY
var torsoGeometry = makeCube();
var scale_torso = new THREE.Matrix4().set(4, 0, 0, 0, 0, 4, 0, 0, 0, 0, 8, 0, 0, 0, 0, 1);
torsoGeometry.applyMatrix(scale_torso);

var headGeometry = makeCube();
var scale_head = new THREE.Matrix4().set(2.5, 0, 0, 0, 0, 2.5, 0, 0, 0, 0, 2.8, 0, 0, 0, 0, 1);
headGeometry.applyMatrix(scale_head);

var tailGeometry = makeCube();
var scale_tail = new THREE.Matrix4().set(0.5, 0, 0, 0, 0, 0.5, 0, 0, 0, 0, 6, 0, 0, 0, 0, 1);
tailGeometry.applyMatrix(scale_tail);

var noseGeometry = makeCube();
var scale_nose = new THREE.Matrix4().set(1.2, 0, 0, 0, 0, 1.2, 0, 0, 0, 0, 1.2, 0, 0, 0, 0, 1);
noseGeometry.applyMatrix(scale_nose);

var stGeometry = makeCube();
var scale_st = new THREE.Matrix4().set(0.4, 0, 0, 0, 0, 0.2, 0, 0, 0, 0, 1.5, 0, 0, 0, 0, 1);
stGeometry.applyMatrix(scale_st);

var ltGeometry = makeCube();
var scale_lt = new THREE.Matrix4().set(0.4, 0, 0, 0, 0, 0.2, 0, 0, 0, 0, 2.5, 0, 0, 0, 0, 1);
ltGeometry.applyMatrix(scale_lt);

var pawGeometry = makeCube();
var scale_paw = new THREE.Matrix4().set(2, 0, 0, 0, 0, 0.8, 0, 0, 0, 0, 3, 0, 0, 0, 0, 1);
pawGeometry.applyMatrix(scale_paw);

var clawGeometry = makeCube();
var scale_claw = new THREE.Matrix4().set(0.2, 0, 0, 0, 0, 0.4, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);
clawGeometry.applyMatrix(scale_claw);


var beamGeometry = makeCylinder();

var pewGeometry = makeSphere();


// TO-DO: SPECIFY THE REST OF YOUR STAR-NOSE MOLE'S GEOMETRY.
// Note: You will be using transformation matrices to set the shape.
// Note: You are not allowed to use the tools Three.js provides for
//       rotation, translation and scaling.
// Note: The torso has been done for you (but feel free to modify it!)
// Hint: Explicity declare new matrices using Matrix4().set

// MATRICES

var tr;
var trt;

var torsoMatrix = new THREE.Matrix4().set(1, 0, 0, 0, 0, 1, 0, 2.5, 0, 0, 1, 0, 0, 0, 0, 1);
var finalTorsoMatrix = torsoMatrix;
var torsoRotMatrix = torsoMatrix;

var headMatrix = new THREE.Matrix4().set(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 4.7, 0, 0, 0, 1);
var finalHeadMatrix = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, headMatrix);
var headRotMatrix = headMatrix;

var noseMatrix = new THREE.Matrix4().set(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 2, 0, 0, 0, 1);
var finalNoseMatrix = new THREE.Matrix4().multiplyMatrices(finalHeadMatrix, noseMatrix);
var noseRotMatrix = noseMatrix;

var tentacleMtxs = [];
var finalTentacleMtxs = [];
for (i = 0; i < 22; i++) {
    var tPosMtx = new THREE.Matrix4().set(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1.4, 0, 0, 0, 1);
    var translate = new THREE.Matrix4().set(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, -1, 0, 0, 0, 1);
    var translateBack = new THREE.Matrix4().set(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1);
    var rotateUp = new THREE.Matrix4().set(1, 0, 0, 0,
        0, Math.cos(-Math.PI / 2), -Math.sin(-Math.PI / 2), 0,
        0, Math.sin(-Math.PI / 2), Math.cos(-Math.PI / 2), 0,
        0, 0, 0, 1);
    var rotatePan = new THREE.Matrix4().set(Math.cos(-2 * (i - 0.5) * Math.PI / 22), 0, Math.sin(-2 * (i - 0.5) * Math.PI / 22), 0,
        0, 1, 0, 0,
        -Math.sin(-2 * (i - 0.5) * Math.PI / 22), 0, Math.cos(-2 * (i - 0.5) * Math.PI / 22), 0,
        0, 0, 0, 1);
    var tilt = new THREE.Matrix4().set(1, 0, 0, 0,
        0, Math.cos(Math.PI / 4), -Math.sin(Math.PI / 4), 0,
        0, Math.sin(Math.PI / 4), Math.cos(Math.PI / 4), 0,
        0, 0, 0, 1);
    var rr = new THREE.Matrix4().multiplyMatrices(rotateUp, rotatePan);
    rr.multiplyMatrices(rr, tilt);

    tr = new THREE.Matrix4().multiplyMatrices(translate, rr);
    trt = new THREE.Matrix4().multiplyMatrices(tr, translateBack)
    tentacleMtxs[i] = new THREE.Matrix4().multiplyMatrices(tPosMtx, trt)
    finalTentacleMtxs[i] = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, tentacleMtxs[i]);
}
var tentacleRotMtxs = tentacleMtxs.slice();

var tailMatrix = new THREE.Matrix4().set(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, -6, 0, 0, 0, 1);
var rotate_tail = new THREE.Matrix4().set(1, 0, 0, 0,
    0, Math.cos(-Math.PI / 20), -Math.sin(-Math.PI / 20), 0,
    0, Math.sin(-Math.PI / 20), Math.cos(-Math.PI / 20), 0,
    0, 0, 0, 1);
tailMatrix.multiplyMatrices(tailMatrix, rotate_tail);
var finalTailMatrix = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, tailMatrix);
var tailRotMatrix = tailMatrix;

var pawMatrices = [new THREE.Matrix4().set(1, 0, 0, 2, 0, 1, 0, -2.4, 0, 0, 1, 3.2, 0, 0, 0, 1),
    new THREE.Matrix4().set(1, 0, 0, -2, 0, 1, 0, -2.4, 0, 0, 1, 3.2, 0, 0, 0, 1),
    new THREE.Matrix4().set(1, 0, 0, 2, 0, 1, 0, -2.4, 0, 0, 1, -3.2, 0, 0, 0, 1),
    new THREE.Matrix4().set(1, 0, 0, -2, 0, 1, 0, -2.4, 0, 0, 1, -3.2, 0, 0, 0, 1)
];
var finalPawMtxs = [];
for (i = 0; i < 4; i++) {
    finalPawMtxs[i] = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, pawMatrices[i]);
}
var pawRotMtxs = pawMatrices.slice();

function pawToClaw(i) {
    var j;
    switch (true) {
        case (i < 5):
            j = 0;
            break;
        case (i >= 5 && i < 10):
            j = 1;
            break;
        case (i >= 10 && i < 15):
            j = 2;
            break;
        case (i >= 15 && i < 20):
            j = 3;
            break;
    }
    return j;
}
function clawToPaw(i) {
    var k;
    switch (true) {
        case (i == 0 || i == 5 || i == 10 || i == 15):
            k = 0;
            break;
        case (i == 1 || i == 6 || i == 11 || i == 16):
            k = 1;
            break;
        case (i == 2 || i == 7 || i == 12 || i == 17):
            k = 2;
            break;
        case (i == 3 || i == 8 || i == 13 || i == 18):
            k = 3;
            break;
        case (i == 4 || i == 9 || i == 14 || i == 19):
            k = 4;
            break;
    }
    return k;
}

var clawMatrices = [];
var finalClawMtxs = [];
for (i = 0; i < 20; i++) {
    var j = pawToClaw(i);
    var k = clawToPaw(i);

    clawMatrices[i] = new THREE.Matrix4().set(1, 0, 0, 0.8 - 0.4 * k, 0, 1, 0, 0.15, 0, 0, 1, 1.5, 0, 0, 0, 1);
    finalClawMtxs[i] = new THREE.Matrix4().multiplyMatrices(finalPawMtxs[j], clawMatrices[i]);
}
var clawRotMtxs = clawMatrices.slice();


var beamMatrix = new THREE.Matrix4().set(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1);
var rotate_beam = new THREE.Matrix4().set(1, 0, 0, 0,
    0, Math.cos(-Math.PI / 2), -Math.sin(-Math.PI / 2), 0,
    0, Math.sin(-Math.PI / 2), Math.cos(-Math.PI / 2), 0,
    0, 0, 0, 1);
beamMatrix.multiplyMatrices(beamMatrix, rotate_beam);
var finalBeamMatrix = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, beamMatrix);
var beamRotMatrix = beamMatrix;

var pewMatrix = new THREE.Matrix4().set(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1.5, 0, 0, 0, 1);
var finalPewMatrix = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, pewMatrix);
var pewRotMatrix = pewMatrix;

// TO-DO: INITIALIZE THE REST OF YOUR MATRICES
// Note: Use of parent attribute is not allowed.
// Hint: Keep hierarchies in mind!
// Hint: Play around with the headTorsoMatrix values, what changes in the render? Why?


// CREATE BODY
var torso = new THREE.Mesh(torsoGeometry, bodyMaterial);
torso.setMatrix(finalTorsoMatrix);
scene.add(torso);

var head = new THREE.Mesh(headGeometry, headMaterial);
head.setMatrix(finalHeadMatrix);
scene.add(head);

var nose = new THREE.Mesh(noseGeometry, noseMaterial);
nose.setMatrix(finalNoseMatrix);
scene.add(nose);

var tentacles = [];
for (i = 0; i < 22; i++) {
    if (i == 0 || i == 1 || i == 11 || i == 12) {
        var tentacle = new THREE.Mesh(stGeometry, tentacleMaterial);
    }
    else {
        tentacle = new THREE.Mesh(ltGeometry, tentacleMaterial);
    }
    scene.add(tentacle);
    tentacle.setMatrix(finalTentacleMtxs[i]);
    tentacles.push(tentacle);
}

var tail = new THREE.Mesh(tailGeometry, tailMaterial);
tail.setMatrix(finalTailMatrix);
scene.add(tail);

var paws = [];
for (i = 0; i < 4; i++) {
    var paw = new THREE.Mesh(pawGeometry, pawMaterial);
    scene.add(paw);
    paw.setMatrix(finalPawMtxs[i]);
    paws.push(paw);
}

var claws = [];
for (i = 0; i < 20; i++) {
    var claw = new THREE.Mesh(clawGeometry, clawMaterial);
    scene.add(claw);
    claw.setMatrix(finalClawMtxs[i]);
    claws.push(claw);
}

var beam = new THREE.Mesh(beamGeometry, lolMaterial);
beam.setMatrix(finalBeamMatrix);

var pew = new THREE.Mesh(pewGeometry, lolMaterial);
pew.setMatrix(finalPewMatrix);

// TO-DO: PUT TOGETHER THE REST OF YOUR STAR-NOSED MOLE AND ADD TO THE SCENE!
// Hint: Hint: Add one piece of geometry at a time, then implement the motion for that part.
//             Then you can make sure your hierarchy still works properly after each step.


// APPLY DIFFERENT JUMP CUTS/ANIMATIONS TO DIFFERNET KEYS
// Note: The start of "U" animation has been done for you, you must implement the hiearchy and jumpcut.
// Hint: There are other ways to manipulate and grab clock values!!
// Hint: Check THREE.js clock documenation for ideas.
// Hint: It may help to start with a jumpcut and implement the animation after.
// Hint: Where is updateBody() called?
var clock = new THREE.Clock(true);

var p0; // start position or angle
var p1; // end position or angle
var time_length; // total time of animation
var time_start; // start time of animation
var time_end; // end time of animation
var p; // current frame
var animate = true; // animate?

// function init_animation()
// Initializes parameters and sets animate flag to true.
// Input: start position or angle, end position or angle, and total time of animation.
function init_animation(p_start, p_end, t_length) {
    p0 = p_start;
    p1 = p_end;
    time_length = t_length;
    time_start = clock.getElapsedTime();
    time_end = time_start + time_length;
    return;
}


function torsoMove(rMtx) {
    finalTorsoMatrix = new THREE.Matrix4().multiplyMatrices(torsoMatrix, rMtx);
    torso.setMatrix(finalTorsoMatrix);

    finalHeadMatrix = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, headRotMatrix);
    head.setMatrix(finalHeadMatrix);

    finalNoseMatrix = new THREE.Matrix4().multiplyMatrices(finalHeadMatrix, noseRotMatrix);
    nose.setMatrix(finalNoseMatrix);

    for (k = 0; k < 22; k++) {
        finalTentacleMtxs[k] = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, tentacleRotMtxs[k]);
        tentacles[k].setMatrix(finalTentacleMtxs[k]);
    }

    finalTailMatrix = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, tailRotMatrix);
    tail.setMatrix(finalTailMatrix);

    for (i = 0; i < 4; i++) {
        finalPawMtxs[i] = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, pawRotMtxs[i]);
        paws[i].setMatrix(finalPawMtxs[i]);
    }

    for (j = 0; j < 20; j++) {
        finalClawMtxs[j] = new THREE.Matrix4().multiplyMatrices(finalPawMtxs[pawToClaw(j)], clawRotMtxs[j]);
        claws[j].setMatrix(finalClawMtxs[j]);
    }

    finalBeamMatrix = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, beamRotMatrix);
    beam.setMatrix(finalBeamMatrix);
}

function headMove(translate, rotate, translateback) {
    tr = new THREE.Matrix4().multiplyMatrices(translate, rotate);
    trt = new THREE.Matrix4().multiplyMatrices(tr, translateback);

    headRotMatrix = new THREE.Matrix4().multiplyMatrices(headMatrix, trt);
    finalHeadMatrix = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, headRotMatrix);
    head.setMatrix(finalHeadMatrix);

    finalNoseMatrix = new THREE.Matrix4().multiplyMatrices(finalHeadMatrix, noseRotMatrix);
    nose.setMatrix(finalNoseMatrix);

    for (i = 0; i < 22; i++) {
        finalTentacleMtxs[i] = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, tentacleRotMtxs[i]);
        tentacles[i].setMatrix(finalTentacleMtxs[i]);
    }

    finalBeamMatrix = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, beamRotMatrix);
    beam.setMatrix(finalBeamMatrix);
}
function headMoveS(translate, rotate, translateback) {
    var startPos = new THREE.Matrix4().set(Math.cos(-p1), 0, Math.sin(-p1), 0,
        0, 1, 0, 0,
        -Math.sin(-p1), 0, Math.cos(-p1), 0,
        0, 0, 0, 1);
    headMove(translate, startPos, translateback);

    tr = new THREE.Matrix4().multiplyMatrices(translate, rotate);
    trt = new THREE.Matrix4().multiplyMatrices(tr, translateback);

    var headRotMatrixS = new THREE.Matrix4().multiplyMatrices(headRotMatrix, trt);
    finalHeadMatrix = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, headRotMatrixS);
    head.setMatrix(finalHeadMatrix);

    finalNoseMatrix = new THREE.Matrix4().multiplyMatrices(finalHeadMatrix, noseRotMatrix);
    nose.setMatrix(finalNoseMatrix);

    for (i = 0; i < 22; i++) {
        finalTentacleMtxs[i] = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, tentacleRotMtxs[i]);
        tentacles[i].setMatrix(finalTentacleMtxs[i]);
    }

    finalBeamMatrix = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, beamRotMatrix);
    beam.setMatrix(finalBeamMatrix);
}

function tentacleMove(translate, rotate, translateback) {
    tr = new THREE.Matrix4().multiplyMatrices(translate, rotate);
    trt = new THREE.Matrix4().multiplyMatrices(tr, translateback);

    for (i = 0; i < 22; i++) {
        tentacleRotMtxs[i] = new THREE.Matrix4().multiplyMatrices(tentacleMtxs[i], trt);
        finalTentacleMtxs[i] = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, tentacleRotMtxs[i]);
        tentacles[i].setMatrix(finalTentacleMtxs[i]);
    }
}

function tailMove(translate, rotate, translateback) {
    tr = new THREE.Matrix4().multiplyMatrices(translate, rotate);
    trt = new THREE.Matrix4().multiplyMatrices(tr, translateback);

    tailRotMatrix = new THREE.Matrix4().multiplyMatrices(tailMatrix, trt);
    finalTailMatrix = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, tailRotMatrix);
    tail.setMatrix(finalTailMatrix);
}
function tailMoveS(translate, rotate, translateback) {
    var startPos = new THREE.Matrix4().set(Math.cos(-p1), 0, Math.sin(-p1), 0,
        0, 1, 0, 0,
        -Math.sin(-p1), 0, Math.cos(-p1), 0,
        0, 0, 0, 1);
    tailMove(translate, startPos, translateback);

    tr = new THREE.Matrix4().multiplyMatrices(translate, rotate);
    trt = new THREE.Matrix4().multiplyMatrices(tr, translateback);

    var tailRotMatrixS = new THREE.Matrix4().multiplyMatrices(tailRotMatrix, trt);
    finalTailMatrix = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, tailRotMatrixS);
    tail.setMatrix(finalTailMatrix);
}

function pawMove(translate, rotate, translateback, i) {
    tr = new THREE.Matrix4().multiplyMatrices(translate, rotate);
    trt = new THREE.Matrix4().multiplyMatrices(tr, translateback);

    pawRotMtxs[i] = new THREE.Matrix4().multiplyMatrices(pawMatrices[i], trt);
    finalPawMtxs[i] = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, pawRotMtxs[i]);
    paws[i].setMatrix(finalPawMtxs[i]);

    var j;
    switch (true) {
        case (i == 0):
            j = 0;
            break;
        case (i == 1):
            j = 5;
            break;
        case (i == 2):
            j = 10;
            break;
        case (i == 3):
            j = 15;
            break;
    }

    for (k = j; k < j + 5; k++) {
        clawRotMtxs[k] = new THREE.Matrix4().multiplyMatrices(clawMatrices[k], trt);
        finalClawMtxs[k] = new THREE.Matrix4().multiplyMatrices(finalPawMtxs[i], clawRotMtxs[k]);
        claws[k].setMatrix(finalClawMtxs[k]);
    }
}
function pawMoveS(translate, rotate, translateback, i) {
    var startPos = new THREE.Matrix4().set(1, 0, 0, 0,
        0, Math.cos(p1), -Math.sin(p1), 0,
        0, Math.sin(p1), Math.cos(p1), 0,
        0, 0, 0, 1);
    pawMove(translate, startPos, translateback, i);

    var pawRotMtxsS = [];
    var clawRotMtxsS = [];

    tr = new THREE.Matrix4().multiplyMatrices(translate, rotate);
    trt = new THREE.Matrix4().multiplyMatrices(tr, translateback);

    pawRotMtxsS[i] = new THREE.Matrix4().multiplyMatrices(pawRotMtxs[i], trt);
    finalPawMtxs[i] = new THREE.Matrix4().multiplyMatrices(finalTorsoMatrix, pawRotMtxsS[i]);
    paws[i].setMatrix(finalPawMtxs[i]);

    var j;
    switch (true) {
        case (i == 0):
            j = 0;
            break;
        case (i == 1):
            j = 5;
            break;
        case (i == 2):
            j = 10;
            break;
        case (i == 3):
            j = 15;
            break;
    }

    for (k = j; k < j + 5; k++) {
        clawRotMtxsS[k] = new THREE.Matrix4().multiplyMatrices(clawRotMtxs[k], trt);
        finalClawMtxs[k] = new THREE.Matrix4().multiplyMatrices(finalPawMtxs[i], clawRotMtxsS[k]);
        claws[k].setMatrix(finalClawMtxs[k]);
    }
}

function beamScale(translate, scale, translateback) {
    tr = new THREE.Matrix4().multiplyMatrices(translate, scale);
    trt = new THREE.Matrix4().multiplyMatrices(tr, translateback);
    var translatePB = new THREE.Matrix4().set(1, 0, 0, 0,
        0, 1, 0, -0.5,
        0, 0, 1, 0,
        0, 0, 0, 1);

    var beamScaleMatrix = new THREE.Matrix4().multiplyMatrices(beamMatrix, trt);
    beamRotMatrix = new THREE.Matrix4().multiplyMatrices(beamScaleMatrix, translatePB);
    finalBeamMatrix = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, beamRotMatrix);
    beam.setMatrix(finalBeamMatrix);
}

function pewpew(projectory) {
    pewRotMatrix = new THREE.Matrix4().multiplyMatrices(pewMatrix, projectory);
    finalPewMatrix = new THREE.Matrix4().multiplyMatrices(finalNoseMatrix, pewRotMatrix);
    pew.setMatrix(finalPewMatrix);
    ;
}

function updateBody() {
    var time = clock.getElapsedTime();

    if (time > time_end || !animate) {
        p = p1;
    }
    else {
        p = (p1 - p0) * ((time - time_start) / time_length) + p0; // current frame
    }

    var rotateX = new THREE.Matrix4().set(Math.cos(-p), 0, Math.sin(-p), 0,
        0, 1, 0, 0,
        -Math.sin(-p), 0, Math.cos(-p), 0,
        0, 0, 0, 1);
    var rotateXB = new THREE.Matrix4().set(Math.cos(p), 0, Math.sin(p), 0,
        0, 1, 0, 0,
        -Math.sin(p), 0, Math.cos(p), 0,
        0, 0, 0, 1);
    var rotateXS = new THREE.Matrix4().set(Math.cos(p * 2), 0, Math.sin(p * 2), 0,
        0, 1, 0, 0,
        -Math.sin(p * 2), 0, Math.cos(p * 2), 0,
        0, 0, 0, 1);
    var rotateZ = new THREE.Matrix4().set(1, 0, 0, 0,
        0, Math.cos(-p), -Math.sin(-p), 0,
        0, Math.sin(-p), Math.cos(-p), 0,
        0, 0, 0, 1);
    var rotateZB = new THREE.Matrix4().set(1, 0, 0, 0,
        0, Math.cos(p), -Math.sin(p), 0,
        0, Math.sin(p), Math.cos(p), 0,
        0, 0, 0, 1);
    var translateH = new THREE.Matrix4().set(1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, -1,
        0, 0, 0, 1);
    var translateHB = new THREE.Matrix4().set(1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, 1,
        0, 0, 0, 1);
    var translateT = new THREE.Matrix4().set(1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, 2.4,
        0, 0, 0, 1);
    var translateTB = new THREE.Matrix4().set(1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, -2.4,
        0, 0, 0, 1);
    var translateP = new THREE.Matrix4().set(1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, -1,
        0, 0, 0, 1);
    var translatePB = new THREE.Matrix4().set(1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, 1,
        0, 0, 0, 1);
    var reset = new THREE.Matrix4().set(1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, 0,
        0, 0, 0, 1);

    var scaleB = new THREE.Matrix4().set(8 * p, 0, 0, 0,
        0, 1000 * p, 0, 0,
        0, 0, 1, 0,
        0, 0, 0, 1);
    var projectory = new THREE.Matrix4().set(1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, 10000 * p,
        0, 0, 0, 1);

    switch (true) {

        case(key == "U"):
            torsoMove(rotateZ);
            break;

        case(key == "E"):
            torsoMove(rotateZB);
            break;

        case(key == "H"):
            headMove(translateH, rotateX, translateHB);
            break;

        case(key == "G"):
            headMove(translateH, rotateXB, translateHB);
            break;

        case(key == "T"):
            tailMove(translateT, rotateX, translateTB);
            break;

        case(key == "V"):
            tailMove(translateT, rotateXB, translateTB);
            break;

        case(key == "N"):
            tentacleMove(translateH, rotateZ, translateHB);
            break;

        case(key == "D"):
            pawMove(translateP, rotateZB, translatePB, 0);
            pawMove(translateP, rotateZB, translatePB, 1);
            pawMove(translateP, reset, translatePB, 2);
            pawMove(translateP, reset, translatePB, 3);
            break;

        case(key == "S" && ss == 0):
            pawMove(translateP, rotateZB, translatePB, 0);
            pawMove(translateP, rotateZB, translatePB, 3);
            pawMove(translateP, reset, translatePB, 1);
            pawMove(translateP, reset, translatePB, 2);
            tailMove(translateT, rotateX, translateTB);
            headMove(translateH, rotateX, translateHB);
            break;

        case(key == "S" && ss == 1):
            pawMoveS(translateP, rotateZ, translatePB, 0);
            pawMoveS(translateP, rotateZ, translatePB, 3);
            pawMove(translateP, rotateZB, translatePB, 1);
            pawMove(translateP, rotateZB, translatePB, 2);
            tailMoveS(translateT, rotateXS, translateTB);
            headMoveS(translateH, rotateXS, translateHB);
            break;

        case(key == "S" && ss == 2):
            pawMove(translateP, rotateZB, translatePB, 1);
            pawMove(translateP, rotateZB, translatePB, 2);
            pawMove(translateP, reset, translatePB, 0);
            pawMove(translateP, reset, translatePB, 3);
            tailMove(translateT, rotateXB, translateTB);
            headMove(translateH, rotateXB, translateHB);
            break;

        case(key == "F"):
            tentacleMove(translateH, rotateZ, translateHB);
            beamScale(translateH, scaleB, translateHB);
            scene.add(beam);
            break;

        case(key == "X"):
            scene.add(pew);
            pewpew(projectory);
            break;






        // TO-DO: IMPLEMENT JUMPCUT/ANIMATION FOR EACH KEY!
        // Note: Remember spacebar sets jumpcut/animate!


        default:
            break;
    }

}


// LISTEN TO KEYBOARD
// Hint: Pay careful attention to how the keys already specified work!
var keyboard = new THREEx.KeyboardState();
var grid_state = false;
var key;

var uFlag, eFlag = false;
var hFlag, gFlag = false;
var nFlag = false;
var tFlag, vFlag = false;
var dFlag = false;
var sFlag = 0;
var ss = 0;

var bFlag = false;

function keyMove(k, flag, f0, f1) {
    if (!flag) {
        (init_animation(f0, f1, 1), key = k)
    }
    else {
        (init_animation(f1, f0, 1), key = k)
    }
    return !flag;
}

function keySwim(sFlag, f0, f1) {
    if (sFlag == 0) {
        ss = 0;
        (init_animation(f0, f1, 1), key = "S")
        return 1;
    }
    else if (sFlag == 1) {
        ss = 1;
        (init_animation(f0, f1, 1), key = "S")
        return 2;
    }
    else {
        ss = 2;
        (init_animation(f1, f0, 1), key = "S")
        return 0;
    }
}


function fireBeam(flag, f0, f1) {
    if (!flag) {
        (init_animation(f0, f1, 0.1), key = "F")
    }
    else {
        (init_animation(f1, f0, 0.1), key = "F")
    }
    return !flag;
}

function firePew(flag, f0, f1) {
    if (!flag) {
        (init_animation(f0, f1, 1), key = "N")
    }
    else {
        (init_animation(f0, f1, 20), key = "X")
    }
    return true;
}

keyboard.domElement.addEventListener('keydown', function (event) {

    if (event.repeat)
        return;
    if (keyboard.eventMatches(event, "Z")) {  // Z: Reveal/Hide helper grid
        grid_state = !grid_state;
        grid_state ? scene.add(grid) : scene.remove(grid);
    }
    else if (keyboard.eventMatches(event, "0")) {    // 0: Set camera to neutral position, view reset
        camera.position.set(45, 0, 0);
        camera.lookAt(scene.position);
    }
    else if (keyboard.eventMatches(event, "U")) {
        uFlag = keyMove("U", uFlag, 0, Math.PI / 4);
        eFlag = false;
    }
    else if (keyboard.eventMatches(event, "E")) {
        eFlag = keyMove("E", eFlag, 0, Math.PI / 4);
        uFlag = false;
    }
    else if (keyboard.eventMatches(event, "H")) {
        hFlag = keyMove("H", hFlag, 0, Math.PI / 6);
        gFlag = false;
    }
    else if (keyboard.eventMatches(event, "G")) {
        gFlag = keyMove("G", gFlag, 0, Math.PI / 6);
        hFlag = false;
    }
    else if (keyboard.eventMatches(event, "T")) {
        tFlag = keyMove("T", tFlag, 0, Math.PI / 6);
        vFlag = false;
    }
    else if (keyboard.eventMatches(event, "V")) {
        vFlag = keyMove("V", vFlag, 0, Math.PI / 6);
        tFlag = false;
    }
    else if (keyboard.eventMatches(event, "N")) {
        scene.remove(beam);
        nFlag = keyMove("N", nFlag, 0, Math.PI / 4);
        bFlag = false;
    }
    else if (keyboard.eventMatches(event, "D")) {
        dFlag = keyMove("D", dFlag, 0, Math.PI / 6);
        sFlag = 0;
    }
    else if (keyboard.eventMatches(event, "S")) {
        sFlag = keySwim(sFlag, 0, Math.PI / 6);
        dFlag = false;
    }
    else if (keyboard.eventMatches(event, "space")) {
        animate = !animate;
    }

    else if (keyboard.eventMatches(event, "F")) {
        bFlag = fireBeam(bFlag, 0, Math.PI / 4);
        nFlag = bFlag;
    }
    else if (keyboard.eventMatches(event, "X")) {
        nFlag = firePew(nFlag, 0, Math.PI / 4);
    }


    // TO-DO: BIND KEYS TO YOUR JUMP CUTS AND ANIMATIONS
    // Note: Remember spacebar sets jumpcut/animate!
    // Hint: Look up "threex.keyboardstate by Jerome Tienne" for more info.


});

// SETUP UPDATE CALL-BACK
// Hint: It is useful to understand what is being updated here, the effect, and why.
function update() {
    updateBody();

    requestAnimationFrame(update);
    renderer.render(scene, camera);
}

update();